/*
 * Font stuff.
 */

#include <stdlib.h>
#include <stdio.h>
#include <vga.h>
#include <vgagl.h>
#include <vgamouse.h>
#include <string.h>
#include <unistd.h>
#include <gif_lib.h>

#include "hr_defines.h"
#include "hr_buttons.h"
#include "hr_general.h"
#include "hr_bitmaps.h"
#include "hr_strings.h"
#include "hr_gif.h"

struct hr_font{
   int flags;
   int index;
   unsigned char *bitmap;
   unsigned char *in_use[32];
   int offset[256];
   int width;
   int height;
#if 0
   int iwidth[256];
   int iheight[256];
#endif
   int scaled_width;
   int scaled_height;
   int x_offset;
   int y_offset;
	int rotate_angle;
   double italics_factor;
   double x_multiplier;
   double y_multiplier;
};


/* Creates a fixed-width font.  Font must be in 16*16 matrix, since its width
 * and height are calculated run-time. Fonts are stored
 * one at a time, in row-major order.
 * Returns NULL if an error occurs.*/
struct hr_font *hr_create_font(struct hr_bitmap *bmap,int index,int flags)
{
	struct hr_font *font=NULL;
	int c,x,y,offset,offset2;
	
	if(bmap!=NULL){
		/* Malloc font struct */
		font=malloc(sizeof(struct hr_font));
		font->flags=flags;
		font->index=index;
		font->width=(bmap->width)/16;
		font->height=(bmap->height)/16;
#if 0
		/* Copy the width data to individual characters. This is the default.*/
		for(c=0;c<256;c++){
			font->iwidth[c]=font->width;
			font->iheight[c]=font->height;
		}
#endif
		font->scaled_width=FONT_DEFAULT_SCALED_WIDTH;
		font->scaled_height=FONT_DEFAULT_SCALED_HEIGHT;
		font->x_offset=FONT_DEFAULT_X_OFFSET;
		font->y_offset=FONT_DEFAULT_Y_OFFSET;
		font->rotate_angle=FONT_DEFAULT_ROTATE_ANGLE;
		font->italics_factor=FONT_DEFAULT_ITALICS_FACTOR;
		font->x_multiplier=FONT_DEFAULT_X_MULTIPLIER;
		font->y_multiplier=FONT_DEFAULT_X_MULTIPLIER;
		
		/* calloc zeroed space for font bitmap. */
		if(flags & FONT_MONO)
		  font->bitmap=calloc(1+((bmap->width)*(bmap->height))/8,1);
		else if (flags & FONT_COLOR)
		  font->bitmap=calloc((bmap->width)*(bmap->height),1);
		
		/* Convert the picture into a font.*/
		for(c=0;c<256;c++){
			for(y=0;y<(font->height);y++){
				for(x=0;x<(font->width);x++){
					/* Calc position of (topleft pixel of the character)
					 * inside bmap.*/
					offset=((c/16)*(font->height)*(bmap->width)+
							  (c%16)*(font->width));
					/* Add Y- and X-offsets within character.*/
					offset+=y*(bmap->width)+x;
					/* Now set corresponding bit or byte in font bitmap.
					 * First calc offset (in bits or bytes).*/
					offset2=c*(font->width)*(font->height)+y*(font->width)+x;
					/* Mono font. */
					if(flags & FONT_MONO){
						if(*((bmap->data)+offset/8) & (128>>(offset%8)))
						  *((font->bitmap)+offset2/8) |= (128>>(offset2%8));
					}
					/* 256color font.*/
					else if(flags & FONT_COLOR){
						if(*(bmap->data+offset)>0)
						  *((font->bitmap)+offset2)=*((bmap->data)+offset);
					}					
				}
			}
		}
		/* Create offset table.*/
		for(c=0;c<256;c++)
		  font->offset[c]=c*(font->width)*(font->height);
		hr_delete_bitmap(bmap);
	}
	return font;
}


/* Writes a character, or part of it.
 * Color 0 is not written.*/
/* Returns number of pixels used horizontally. */
int hr_partial_write(int x,int y,int color,int bcolor,
							 unsigned char character,
							 int px1,int py1,int px2,int py2,
							 struct hr_font *font,int flags)
{
	int x2,y2,xm=0,byte_offset,bit_offset;
	int character_offset,pixel_offset,pixcolor;
	if(font!=NULL) {
		/* Exit if any of these is out of range. FIX! Check FONT_PROPORTIONAL.*/
		if((px1<0)||(py1<0)||(px2>=(font->width))||(py2>=(font->height)))
		  return 0;
		/* Get offset of the 1st pixel in this character.*/
		character_offset=font->offset[character];
		/* Let's make things easier */
		x-=px1;y-=py1;
		/* Scan the bits...*/
		for(y2=py1;y2<=py2;y2++) {
			/* Pixel offset must be calced more often than in plain
			 * write() because not every pixel is drawn horizontally.*/
			pixel_offset=character_offset+(y2*(font->width))+px1;
			for(x2=px1;x2<=px2;x2++) {
				pixcolor=-1;
				/* "black & white" fonts.*/
				if(font->flags & FONT_MONO) {
					/* Now divide pixel offset into bytes and bits.*/
					byte_offset=pixel_offset/8;
					bit_offset =pixel_offset%8;
					/* If the bit is set, draw one pixel UNLESS
					 * FONT_REVERSE is set... FIX! Cleanup...*/
					if(*((font->bitmap)+byte_offset) & (256>>bit_offset)){
						if(flags & FONT_REVERSE){
							if(flags & FONT_SINGLECOLBGR)
							  pixcolor=bcolor;
						}
						else
						  pixcolor=color;
					}
					else{
						if(flags & FONT_REVERSE)
						  pixcolor=color;
						else if(flags & FONT_SINGLECOLBGR)
						  pixcolor=bcolor;
					}
				}
				/* Colored fonts. */
				else if(font->flags & FONT_COLOR) {
					/* Pixel offset is now directly the offset in bytes.
					 * Color is taken from the table, and a pixel is drawn
					 * if color is non-zero.*/
					if(*((font->bitmap)+pixel_offset)){
						/* FONT_REVERSE works strangely with 256 color fonts.*/
						if(flags & FONT_REVERSE){
							if(flags & FONT_SINGLECOLBGR)
							  pixcolor=bcolor;
						}
						else
						  pixcolor=*((font->bitmap)+pixel_offset)+color;
					}
					else{
						if(flags & FONT_REVERSE)
						  pixcolor=color;
						else if(flags & FONT_SINGLECOLBGR)
						  pixcolor=bcolor;
					}
				}
				/* Set the X modifier if FONT_ITALICS is set.
				 * Note that the tilt angle cannot be changed within a string.
				 * If you must use the same font tilted in many different
				 * angles, call hr_write many times and change font->italics_factor
				 * in between hr_write() calls.*/
				if(flags & FONT_ITALICS)
				  xm=(int)(((double)(font->height)/2.0-(double)(y2))*
							  (font->italics_factor)+0.5);
				/* FONT_INVISIBLE is useful for finding out how much
				 * space WOULD be used, had the string been printed.*/
				if(!(flags & FONT_INVISIBLE)){
					/* Draw the pixel (finally) */
					if(pixcolor>=0){
						gl_setpixel(x+x2+xm,y+y2,pixcolor);
						if(flags & FONT_BOLD)
						  gl_setpixel(x+x2+xm+1,y+y2,pixcolor);
					}
					/* Draw underline, if necessary. */
					else if((flags & FONT_UNDERLINED) && (y2==(font->height)-1)){
						gl_setpixel(x+x2,y+y2,color);
					}
				}
				/* Move pixel_offset to next pixel within the character.*/
				++pixel_offset;
			}
		}
		/* Return number of pixels used.
		 * NOTE! x/y_multipliers are NOT taken into account.
		 * They are meant just for text flipping and should always be
		 * -1<<1. */
#if 0
		if(flags & FONT_PROPORTIONAL)
		  return font->iwidth[character];
		else
#endif
		  if(flags & FONT_SCALED)
		  return font->scaled_width;
		return font->width;
	}
	return 0;
}


/* Writes a string. It is possible to contain formatting codes within
 * a string. Just add a FONT_TOGGLECHAR, followed by
 * b - Bold
 * u - Underlined
 * r - Inverse
 * s - Single color background
 * i - Italics
 * z - Scaling
 * x - X multiplication
 * y - Y multiplication
 * o - Rotated
 * p - Proportional
 * 0-9 - Set italics_factor. '0'=90 degrees, '4'=45 degrees
 * t   - Print the TOGGLECHAR.
 * 
 * f   - Print the FORMATCHAR.
 * 
 * The func returns the last X coordinate used.
 */
int hr_write(int x,int y,int color,int bcolor,
				 unsigned char *text,
				 struct hr_font *font,int offset,int len,int flags)
{
	int c,curchar,printed_width=0,printed_height=0,printed=0,printables=0;
	/**/
	if((text!=NULL)&&(font!=NULL)) {
		printed_height=(font->height);
		if(flags & FONT_UNFORMATTED){
			/* Scan all characters in a string.*/
			for(c=0;c<strlen(text);c++) {
				/* Character. Not necessarily printable,
				 * but hr_partial_write() checks that. */
				if((printables>=offset)&&((printed<len)||(len==STRING_END))){
					printed_width+=hr_partial_write(x+printed*(font->width),y,
															  color,bcolor,
															  (unsigned char)(*(text+c)),
															  0,0,(font->width)-1,
															  (font->height)-1,
															  font,flags);
					++printed;
				}
				++printables;
			}
		}
		else{
			/* Scan all items in string */
			for(c=0;c<hr_strlen(text);c++) {
				/* Check what each item is.*/
				curchar=hr_string_item(text,c);
				switch(curchar){
				 case STRING_TOGGLESEQ:
					/* TOGGLE sequence.*/
					flags=hr_parse_controlseq(text+hr_string_item_offset(text,c),
													  flags,STRING_RIGHT);
					break;
				 case STRING_FORMATSEQ:
					/* FORMAT sequence.*/
					break;
				 default:
					/* Character. Not necessarily printable,
					 * but hr_partial_write() checks that. */
					if((printables>=offset)&&((printed<len)||(len==STRING_END))){
						printed_width+=hr_partial_write(x+printed*(font->width),y,
																  color,bcolor,
																  (unsigned char)(curchar),
																  0,0,(font->width)-1,
																  (font->height)-1,
																  font,flags);
						++printed;
					}
					++printables;
					break;
				}
			}
		}
		if(flags & FONT_RETURN_HEIGHT)
		  return printed_height;
		else
		  return printed_width;
	}
	return x;
}


/* Frees memory used by a font.*/
struct hr_font *hr_delete_font(struct hr_font *font)
{
	if(font!=NULL){
		free(font->bitmap);
		free(font);
	}
	return font;
}
